# -*- coding: utf-8 -*-
# Copyright 2018 Alex Marchenko
# Distributed under the terms of the Apache License 2.0
"""
Kafka Producer
#################
This module provides a simple interface for generating messages from
:class:`~dronedirector.aerial.Drone`s
about their location (altitude, latitude, longitude). An 
Avro schema is provided for :class:`~dronedirector.aerial.Drone` messages.
If a schema registry is not available, JSON messages can be sent using
:func:`~dronedirector.producer.fly_drones`.
"""
from __future__ import print_function
try:
    import confluent_kafka as ck
except ImportError:
    raise ImportError("Confluent Kafka Python client not available")
try:
    from confluent_kafka import avro
except ImportError:
    print("Only the JSON producer is available -- use fly_drones")
import time


# value schema
drone_schema_str = """
{
    "namespace": "dronedirector.drones",
    "name": "drone",
    "type": "record",
    "fields": [
        {
            "name": "uid",
            "type": "string"
        },
        {
            "name": "region",
            "type": "string"
        },
        {
            "name": "dronetime",
            "type": "string"
        },
        {
            "name": "altitude",
            "type": "double"
        },
        {
            "name": "latitude",
            "type": "double"
        },
        {
            "name": "longitude",
            "type": "double"
        }
    ]
}"""


def fly_avro_drones(bootstrap_servers, schema_registry_url, nmessages, default_value_schema_str=drone_schema_str,
               producer_dict_kwargs=None, topic_name="drones_raw", time_delay=0, drones=None):
    """
    A simple example of sending structured messages from drones to a message broker.

    Args:
        bootstrap_servers (str): Comma separated string of Kafka servers
        schema_registry_url (str): Schema registry URL
        nmessages (int): Number of messages to send
        default_value_schema_str (str): String Avro schema compatible with mdrone messages
        producer_dict_kwargs (dict): Optional keyword arguments for producer
        topic_name (str): Topic name to which drone messages will be sent
        time_delay (int): Delay time between cycles when producing messages
        drones (iterable): Iterable of drones from which to generate messages

    Tip:
        Schemas should match the messages sent by drones.
    """
    pdk = {'bootstrap.servers': bootstrap_servers,
           'schema.registry.url': schema_registry_url},
    if isinstance(producer_dict_kwargs, dict):
        pdk.update(producer_dict_kwargs)
    producer = avro.AvroProducer(pdk, default_value_schema=avro.loads(default_value_schema_str))
    z = len(str(nmessages))            # Pretty print cycle number for logging
    for i in range(nmessages):
        print("====MESSAGE SET {}====".format(str(i).zfill(z)))
        for drone in drones:
            msg = drone.message()
            print(msg)
            producer.produce(topic=topic_name, value={k: getattr(msg, k) for k in msg._fields})
        time.sleep(time_delay)
    producer.flush()


def fly_drones(bootstrap_servers, nmessages, producer_dict_kwargs=None,
               topic_name="drones_raw", time_delay=0, drones=None):
    """
    A simple example of sending drones that send JSON messages to the message broker.

    Args:
        bootstrap_servers (str): Comma separated string of Kafka servers
        nmessages (int): Number of messages to send
        producer_dict_kwargs (dict): Optional keyword arguments for producer
        topic_name (str): Topic name to which drone messages will be sent
        time_delay (int): Delay time between cycles when producing messages
        drones (iterable): Iterable of drones from which to generate messages
    """
    pdk = {'bootstrap.servers': bootstrap_servers}
    if isinstance(producer_dict_kwargs, dict):
        pdk.update(producer_dict_kwargs)
    producer = ck.Producer(pdk)
    z = len(str(nmessages))            # Pretty print cycle number for logging
    for i in range(nmessages):
        print("====MESSAGE SET {}====".format(str(i).zfill(z)))
        for drone in drones:
            msg = drone.message()
            print(msg)
            producer.produce(topic_name, msg)
        time.sleep(time_delay)
    producer.flush()



# -*- coding: utf-8 -*-
# Copyright 2018 Alex Marchenko
# Distributed under the terms of the Apache License 2.0
from .aerial import Drone, AerialObject, SinusoidalDrone
from .distance import lat_to_rad, lon_to_rad, earth_radius



# -*- coding: utf-8 -*-
# Copyright 2018 Alex Marchenko
# Distributed under the terms of the Apache License 2.0
"""
Distance Calculation Utilities
################################
This module provides functions for determining distances between aerial
objects (static and dynamic) which are near the surface of planet Earth.
This module makes use of compilation to speed up mathematical operations.

Tip:
    Altitudes are expected in meters, latitudes and longitudes are expected
    in degrees. By convention N and E are positive angles and S and W are
    negative angles.

See Also:
    https://en.wikipedia.org/wiki/World_Geodetic_System
"""
from __future__ import absolute_import, division
import numpy as np
import numba as nb


equatorial_radius = 6378137.0               # Equatorial radius in m
polar_radius = 6356752.3                    # Polar radius in m
_a = equatorial_radius
_a2 = _a**2
_b = polar_radius
_b2 = _b**2


@nb.jit(nopython=True, nogil=True)
def lat_to_rad(degree):
    """
    Convert latitude (in degrees) to radians.

    Args:
        degree (float): Angle in degrees (e.g. +41.0 == 41.0 N).

    Returns:
        rad (float): Angle in radians
    """
    if abs(degree) > 90:
        raise ValueError("Latitudes span the range -90 <= degree <= +90")
    return degree*np.pi/180


@nb.jit(nopython=True, nogil=True)
def lon_to_rad(degree):
    """
    Convert longitude (in degrees) to radians.

    Args:
        degree (float): Angle in degrees (e.g. -122.0 == 122.0 W).

    Returns:
        rad (float): Angle in radians
    """
    if abs(degree) > 180:
        raise ValueError("Longitudes span the range -180 <= degree <= +180")
    return degree*np.pi/180

    
#@nb.jit(nopython=True, nogil=True)
def earth_radius(lat):
    """
    Compute the radius of the Earth for a given latitude.
    
    The Earth isn't exactly spherical; depending on what latitude
    an object is at, the distance from sea-level to the center of
    the Earth varies slightly.

    Args:
        lat (float): Latitude in radians

    Returns:
        radius (float): Approximate distance to sea level at the current latitude in meters
    """
    return np.sqrt(((_a2*np.cos(lat))**2 + (_b2*np.sin(lat))**2)/
                   ((_a*np.cos(lat))**2 + (_b*np.sin(lat))**2))
    
    
@nb.jit(nopython=True, nogil=True)
def to_cartesian(alt, lat, lon):
    """
    Convert altitude, latitude, and longitude to spherical coordinates.

    Distances are computed relative to the center of the earth.
    
    Args:
        alt (float): Altitude (above sea-level) in meters
        lat (float): Latitude in radians
        lon (float): Longitude in radians

    Returns:
        tup (tuple): Tuple of x, y, z coordinates in meters
    """
    r = alt + earth_radius(lat)
    x = r*np.sin(lat)*np.cos(lon)
    y = r*np.sin(lat)*np.sin(lon)
    z = r*np.cos(lat)
    return x, y, z, r

@nb.jit(nopython=True, nogil=True)
def compute_distances(xs, ys, zs):
    """
    Brute force computation of pairwise distances.

    Compute the distance between one object represented by the cartesian coordinates
    (x, y, z) and a collection of objects represented by arrays {xs}, {ys}, {zs}.
    
    Args:
    xs (array): Cartesian x coordinates of target objects
    ys (array): Cartesian y coordinates of target objects
    zs (array): Cartesian z coordinates of target objects
    """
    n = len(xs)
    dxyz = np.empty((n*(n-1)//2, 4))
    k = 0
    for i in range(n):
        for j in range(i+1, n):
            dx = (x - xs[i])
            dy = (y - ys[i])
            dz = (z - zs[i])
            dr = np.sqrt(dx**2 + dy**2 + dz**2)
            dxyz[k, 0] = dx
            dxyz[k, 1] = dy
            dxyz[k, 2] = dz
            dxyz[k, 3] = dr
    return dxyz



# -*- coding: utf-8 -*-
# Copyright 2018 Alex Marchenko
# Distributed under the terms of the Apache License 2.0
"""
Aerial Objects
#################
This module provides some simple classes that are used to simulate flying
drones and other objects.

Tip:
    Altitudes are expected in meters, latitudes and longitudes are expected
    in degrees. By convention N and E are positive angles and S and W are
    negative angles.

See Also:
    https://en.wikipedia.org/wiki/World_Geodetic_System
"""
from __future__ import absolute_import
import six
import uuid
import json
import types
import numpy as np
from datetime import datetime
from itertools import cycle


dtfmt = "%Y%m%d%H%M%S%f"


class AerialObject(object):
    """
    Base class for aerial objects.

    Aerial objects can be static or dynamic but must define something
    about their location (latitude, longitude) and how tall they are/how
    high up they fly. An example of a static aerial object (like a tree
    or building) is given below. Note that at a minimum,altitude, latitude,
    and longitude should be defined.

    .. code:: Python

        class CaliRedwood(AerialObject):
            # Example of a completely static object
            def __init__(self):
                super(CaliRedwood, self).__init__(altitude=itertools.cycle([100.0]),
                                                  latitude=itertools.cycle([37.8716]),
                                                  longitude=itertools.cycle([-122.2727]))
    """
    def message(self):
        """
        Get the current coordinates of the object as a JSON message.

        Returns:
            text (str): JSON formatted message
        """
        return json.dumps(dict(altitude=next(self.altitude),
                               latitude=next(self.latitude),
                               longitude=next(self.longitude)))

    def __init__(self, altitude, latitude, longitude):
        if not all(isinstance(i, (types.GeneratorType, cycle)) for i in (altitude, latitude, longitude)):
            raise TypeError("Altitude, latitude, and longitude must be generators")
        self.altitude = altitude
        self.latitude = latitude
        self.longitude = longitude


class Drone(AerialObject):
    """
    Base class for drones.

    .. code:: Python

        drone = Drone(altitude=itertools.cycle([1000.0]),
                      latitude=itertools.cycle([41.0]),
                      longitude=itertools.cycle(np.sin(np.arange(0, 2*np.pi, np.pi/(360*50)))),
                      region="New York County")

    Args:
        altitude: Generator that yields altitude(s)
        latitude: Generator that yields latitude(s)
        longitude: Generator that yields longitude(s)
        region: Region (coarse location) identifier
        uid: Unique identifier
    """
    def message(self):
        """
        Get the current coordinates of the object as a JSON message.

        Returns:
            text (str): JSON formatted message
        """
        return json.dumps(dict(altitude=next(self.altitude),
                               latitude=next(self.latitude),
                               longitude=next(self.longitude),
                               uid=self.uid.hex,
                               region=self.region,
                               dronetime=datetime.now().strftime(dtfmt)))

    def __init__(self, altitude, latitude, longitude, region, uid=None):
        super(Drone, self).__init__(altitude, latitude, longitude)
        self.uid = uuid.uuid4() if uid is None else uid
        self.region = region


class SinusoidalDrone(Drone):
    """
    A drone that circumnavigates the Earth sinusoidally at a given altitude
    and latitude and with a given angular speed.

    Args:
        alt (float): Drone's constant altitude
        lat (float): Drone's constant latitude
        region: Region (coarse location) identifier
        uid: Unique identifier
        speed (float): Fraction of a (longitudinal) degree moved per second
    """
    def __init__(self, alt, lat, region, uid=None, speed=0.02):
        altitude = cycle([alt])
        latitude = cycle([lat])
        longitude = cycle(np.sin(np.arange(0, 2*np.pi, np.pi/360*speed)))
        super(SinusoidalDrone, self).__init__(altitude=altitude, latitude=latitude,
                                              longitude=longitude, uid=uid, region=region)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
from distutils import log
from setuptools import setup, find_packages


name = "dronedirector"
description = "A simple Python package for simulating aerial drones"
readme = "README.md"
requirements = "requirements.txt"
verfile = "_version.py"
root = os.path.dirname(os.path.abspath(__file__))
log.set_verbosity(log.DEBUG)
try:
    import pypandoc
    long_description = pypandoc.convert(readme, "rst")
except ImportError:
    with open(readme) as f:
        long_description = f.read()
with open(requirements) as f:
    dependencies = f.read().splitlines()
with open(os.path.join(root, name, verfile)) as f:
    v = f.readlines()[-2]
    v = v.split('=')[1].strip()[1:-1]
    version = '.'.join(v.replace(" ", "").split(","))


setup_args = {
    'name': name,
    'version': version,
    'description': description,
    'long_description': long_description,
    'install_requires': dependencies,
    'packages': find_packages(),
    'license': "Apache License Version 2.0",
    'author': "Alex Marchenko",
    'author_email': "alexvmarch@gmail.com",
    'maintainer_email': "alexvmarch@gmail.com",
    'url': "https://github.com/avmarchenko/" + name,
    'download_url': "https://github.com/avmarchenko/{}/archive/v{}.tar.gz".format(name, version)
}

setup(**setup_args)
